import socket


def resolve_hostnames(l):
    r = list()
    assert isinstance(l, (list, set))
    for line in l:
        r.append(socket.gethostbyname(line.strip()))
    return r


def dList2Dict(l):
    return {v[0]: v[1] for v in l}


def dict2dList(l):
    return [[k, v] for k, v in l.items()]



from .trigger import Trigger
from ..log import log
from .. import exceptions


class CancelTrigger(Trigger):
    """ The in progress trigger sets an event to in_play as soon as it starts
    """

    def _trigger(self, args):
        """ Set a BMG to ``in_progress``
        """
        log.info("Cancel event ...")
        event = self.getEvent()
        event.status_update("canceled")

    def testThreshold(self):
        """ The threshold that needs to be crossed in order to cancel an event
            on chain.

            .. alert:: This is temporary set to be ``2`` until we have an
                easier way to identify how many data proxies send data to us
        """
        return 2

    def testConditions(self, *args, **kwargs):
        """ The test conditions for canceling the event are as this:

            * Do more incidents propose the creation of the event than
                ``testThreshold``

               -> cancel the event

        """
        incidents = self.get_all_incidents()
        if not incidents:
            raise exceptions.InsufficientIncidents
        cancel_incidents = incidents.get("canceled", {}).get("incidents")
        if cancel_incidents and len(cancel_incidents) >= self.testThreshold():
            return True
        else:
            log.info(
                "Insufficient incidents for {}({})".format(
                    self.__class__.__name__, str(self.teams)
                )
            )
            raise exceptions.InsufficientIncidents



from ..log import log
from .. import exceptions
from dateutil.parser import parse
from bookied_sync.sport import LookupSport
from bookied_sync.eventgroup import LookupEventGroup
from bookied_sync.event import LookupEvent
from bos_incidents import factory
from bookiesports.normalize import IncidentsNormalizer, NotNormalizableException
from peerplaysapi.exceptions import UnhandledRPCError
from peerplays.bettingmarketgroup import BettingMarketGroups
from peerplays.event import Events
from peerplays.proposal import Proposals


class Trigger:
    """ This class is used to deal with Messages that have been received by any
        means and need processing thru bookied-sync
    """

    def __init__(self, message, lookup_instance, config, clear_caches=True, **kwargs):
        self.message = message
        self.lookup = lookup_instance
        self.config = config

        # Obtain data for unique key
        # The "id" contains everything we need to identify an individual event
        # which itself contains at least the sport, and the teams

        # Get the id (internally used only)
        self.id = message.get("id")

        # Incident Storage
        if "storage" in kwargs and kwargs["storage"]:
            self.storage = kwargs["storage"]
        else:
            self.storage = factory.get_incident_storage(
                kwargs.get("mongodb", None), purge=kwargs.get("purge", False)
            )

        # Normalize incident
        self.normalizer = IncidentsNormalizer(chain=lookup_instance._network_name)
        self.normalize(message)

        # Let's clear the caches for Proposals and events
        # We need to do so because of the internal pypeerplays cache.
        # The cache reduces the API calls to the backend and thus latency.
        # However, to be sure the event hasn't been created since before the
        # cache has expired, we force a refresh from the blockchain.
        if clear_caches:
            Events.clear_cache()
            Proposals.clear_cache()

        # Try obtain the sport
        self.sport = LookupSport(self.id.get("sport"))

        # Given the sport, try to obtain the league (event group)
        self.eventgroup = LookupEventGroup(self.sport, self.id.get("event_group_name"))

        self.event = None  # Will be filled in after receiving a trigger

        # Get Teams from query
        self.teams = [self.id.get("home"), self.id.get("away")]

        # Get start time from query
        self.start_time = parse(self.id.get("start_time", ""))

    @property
    def incident(self):
        """ Return the incident message
        """
        return self.message

    @property
    def call(self):
        """ Return the trigger/call name
        """
        return self.message.get("call").lower()

    def getEvent(self):
        """ Get an event from the lookup
        """
        event = LookupEvent.find_event(
            teams=self.teams,
            start_time=self.start_time,
            eventgroup_identifier=self.eventgroup.identifier,
            sport_identifier=self.sport.identifier,
        )

        if event:

            # Store in object
            self.event = event

            eventgroup = event.eventgroup
            if not eventgroup.is_open:
                log.debug(
                    "Skipping not-yet-open BMG: {}".format(str(eventgroup.identifier))
                )
                raise exceptions.EventGroupClosedException

            return event

        else:
            raise exceptions.EventDoesNotExistException

    def trigger(self, *args, **kwargs):
        """ Forward a trigger to the actual trigger implementation
            in the subclass
        """

        # Test if I am supposed to proceed with this
        self.testConditions()

        # Execute the actual Trigger
        self._trigger(*args, **kwargs)

        # if a proposal is going to be published, let's enable
        # blocking so we can obtain the proposal id
        # FIXME: This can be optimized for speed by putting this into
        # an independent thread or throwing it into the redis queue
        # so the worker can continue with other incidents
        if self.lookup.proposal_operations():
            self.lookup.set_blocking(True)
        else:
            self.lookup.set_blocking(False)

        # Broadcast that stuff
        transactions = self.broadcast()

        # Obtain data from the blockchain
        proposal_ids = [x.get_proposal_id() for x in transactions]
        actions = [x.action() for x in transactions]

        # unless _trigger raises an exception
        self.set_incident_status(
            status_name="done", status_add=dict(proposals=proposal_ids, actions=actions)
        )

        return transactions

    def normalize(self, *args, **kwargs):
        try:
            message = self.normalizer.normalize(self.message, errorIfNotFound=True)
        except NotNormalizableException as e:
            self.set_incident_status(status_name="not normalizable")
            raise e
        if message != self.message:
            try:
                self.storage.delete_incident(self.message)
            except Exception:
                pass
            self.message = message
            self.store_incident()

    def get_all_incidents(self):
        """ Let's get all the incidents for an event
        """
        return self.storage.get_event_by_id(self.message)

    def set_incident_status(self, **kwargs):
        """ We here set the status of an **event** in the incidents storage
        """
        try:
            self.storage.update_event_status_by_id(self.id, call=self.call, **kwargs)
        except Exception:
            pass

    def get_incident_status(self):
        """ Get the current status of an **event** in the incidents storage
        """
        event = self.storage.get_event_by_id(self.id, False)
        return event.get(self.call, {}).get("status", None)

    def broadcast(self):
        """ This method broadcasts the updates to the chain
        """
        """
        try:
        """
        return self.lookup.broadcast()
        """
        except UnhandledRPCError as e:
            if "Proposed operation is already pending for approval" in str(e):
                raise exceptions.ProposalAlreadyExistsOrIsPendingException()
            else:
                raise e
        """

    def store_incident(self):
        """ This call stores the incident in the incident-store (bos-incident)
        """
        self.storage.insert_incident(self.message)

    def get_onchain_bmgs(self):
        """ Returns a list of BettingMarketGroups of the event that already
            exist on the Blockchain
        """
        if not self.event:
            self.event = self.getEvent()
            if not self.event:
                return []
        return BettingMarketGroups(self.event.id)

    # Methods that need to be overwritten by trigger
    def testConditions(self, *args, **kwargs):
        """ Test If we can actually call the trigger. This method is called
            from trigger() and is supposed to be overwritten by the actual
            trigger.
        """
        pass

    def _trigger(self, *args, **kwargs):
        """ To be implemented by the sub class
        """
        pass



from bookied.triggers.create import CreateTrigger
from bookied.triggers.result import ResultTrigger
from bookied.triggers.in_progress import InProgressTrigger
from bookied.triggers.finish import FinishTrigger
from bookied.triggers.cancel import CancelTrigger
from bookied.triggers.dynamic_bmg import DynamicBmgTrigger


INCIDENT_CALLS = [
    "create",
    "in_progress",
    "finish",
    "result",
    "canceled",
    "dynamic_bmgs",
]


TRIGGERS = {
    "create": CreateTrigger,
    "in_progress": InProgressTrigger,
    "finish": FinishTrigger,
    "result": ResultTrigger,
    "canceled": CancelTrigger,
    "dynamic_bmgs": DynamicBmgTrigger,
}



#: Default incident schema as sent from the data proxy
schema = {
    "$schema": "http://json-schema.org/draft-06/schema#",
    "title": "Bookie dataproxy Json",
    "description": "Bookie dataproxy exchange event trigger format",
    "type": "object",
    "properties": {
        "id": {
            "type": "object",
            "properties": {
                "sport": {
                    "description": "The unique name of the sport, in english",
                    "type": "string",
                },
                "event_group_name": {
                    "description": "The unique name of the event group, in english (e.g. the league)",
                    "type": "string",
                },
                "start_time": {
                    "description": "The start time of the event in UTC, ISO format",
                    "type": "string",
                    "format": "date-time",
                },
                "home": {
                    "description": "The unique name of the home team, in english",
                    "type": "string",
                },
                "away": {
                    "description": "The unique name of the away team, in english",
                    "type": "string",
                },
            },
            "required": ["sport", "event_group_name", "start_time", "home", "away"],
        },
        "call": {
            "description": "The trigger that was called",
            "type": "string",
            "enum": ["create", "in_progress", "finish", "result", "unknown", "settle"],
        },
        "arguments": {
            "type": "object",
            "properties": {
                "season": {
                    "description": "The unique season of the sport",
                    "type": "string",
                },
                "whistle_start_time": {
                    "description": "The time the end was whistled on in UTC, ISO format",
                    "type": ["null", "string"],
                    "format": "date-time",
                },
                "whistle_end_time": {
                    "description": "The time the end was whistled off in UTC, ISO format",
                    "type": ["null", "string"],
                    "format": "date-time",
                },
                "home_score": {
                    "description": "The score of the home team",
                    "type": "string",
                },
                "away_score": {
                    "description": "The score of the away team",
                    "type": "string",
                },
            },
        },
    },
    "required": ["id", "call", "arguments"],
}



import logging
from logging.handlers import SMTPHandler, RotatingFileHandler
from colorlog import ColoredFormatter
from .config import loadConfig

HAS_TELEGRAM = False
try:  # pragma: no cover
    import telegram_handler

    HAS_TELEGRAM = True
except Exception:
    pass


# Load config file
config = loadConfig()
logging_config = config.get("logging", {})

# Enable logging
USE_TELEGRAM = (
    bool(HAS_TELEGRAM)
    and logging_config.get("telegram_token", None) is not None
    and logging_config.get("telegram_chatid", None) is not None
)
USE_MAIL = logging_config.get("mailto", None)
USE_STREAM = True  # logging_config.get("console", None)
USE_FILE = logging_config.get("file", None)


def log_stream(logger):  # pragma: no cover
    """ Log content in the output/syslog
    """
    if USE_STREAM:
        formatter = ColoredFormatter(LOGFORMAT)
        stream = logging.StreamHandler()
        stream.setLevel(LOG_LEVEL)
        stream.setFormatter(formatter)
        logger.addHandler(stream)


def log_file(logger):  # pragma: no cover
    """ Log content to a file
    """
    if USE_FILE:
        log_handler_rotate = RotatingFileHandler(
            logging_config.get("file", "bookied.log"),
            maxBytes=1024 * 1024 * 100,
            backupCount=20,
        )
        log_handler_rotate.setLevel(logging.INFO)
        logger.addHandler(log_handler_rotate)


def log_mail(logger):  # pragma: no cover
    """ Send an email for logging
    """
    if USE_MAIL:
        # Mail
        log_handler_mail = SMTPHandler(
            config.get("mailhost", "localhost"),
            config.get("mailfrom", "bookied@example.com"),
            config.get("mailto"),
            config.get("mailsubject", "BookieD notification mail"),
        )
        log_handler_mail.setFormatter(
            logging.Formatter(
                "Message type:       %(levelname)s\n"
                + "Location:           %(pathname)s:%(lineno)d\n"
                + "Module:             %(module)s\n"
                + "Function:           %(funcName)s\n"
                + "Time:               %(asctime)s\n"
                + "\n"
                + "Message:\n"
                + "\n"
                + "%(message)s\n"
            )
        )
        log_handler_mail.setLevel(logging.WARNING)
        logger.addHandler(log_handler_mail)


def log_telegram(logger):  # pragma: no cover
    """ Enable logging via Telegram
    """
    if USE_TELEGRAM:
        tgHandler = telegram_handler.TelegramHandler(
            token=logging_config.get("telegram_token"),
            chat_id=logging_config.get("telegram_chatid"),
        )
        tgHandler.setLevel(logging.WARNING)
        logger.addHandler(tgHandler)


# Default logging facilities
LOG_LEVEL = logging.getLevelName(logging_config.get("level", "INFO"))
LOGFORMAT = (
    "  %(log_color)s%(levelname)-8s%(reset)s |" " %(log_color)s%(message)s%(reset)s"
)
logging.root.setLevel(LOG_LEVEL)
formatter = ColoredFormatter(LOGFORMAT)

log = logging.getLogger("bookied.*")
log.setLevel(LOG_LEVEL)

logsync = logging.getLogger("bookied_sync")
logsync.setLevel(LOG_LEVEL)

for logger in [log, logsync]:
    log_stream(logger)
    log_file(logger)
    log_mail(logger)
    log_telegram(logger)

# Enable full logging
# logging.basicConfig(level=logging.DEBUG)



from redis import Redis
from .config import loadConfig
from .log import log


def get_redis():  # pragma: no cover
    """ Open the redis connection
    """
    config = loadConfig()
    log.info(
        "Opening Redis connection (redis://{}/{})".format(
            config.get("redis_host", "localhost") or "localhost",
            config.get("redis_port", 6379) or 6379,
            config.get("redis_db", 0) or 0,
        )
    )
    return Redis(
        config.get("redis_host", "localhost") or "localhost",
        config.get("redis_port", 6379) or 6379,
        password=config.get("redis_password"),
        db=config.get("redis_db", 0) or 0,
    )



#!/usr/bin/env python3

import threading
import click

from rq import Connection, Worker, use_connection, Queue
from bos_incidents import factory

from . import INCIDENT_CALLS
from .log import log
from .config import loadConfig


config = loadConfig()


@click.group()
def main():
    """ Main group for python-click so we can offer subcommands in a single cli
        tool
    """
    pass


@main.command()
@click.option("--port", type=int, default=8010)
@click.option("--host", type=str, default="127.0.0.1")
@click.option("--debug", is_flag=True, default=False)
@click.option("--proposer")
@click.option("--approver")
@click.option("--scheduler/--no-scheduler", default=True)
def api(port, host, debug, proposer, approver, scheduler):
    """ Start the API endpoint
    """
    if scheduler:
        from .schedule import scheduler as start_scheduler
        from bookied.web import background_threads

        background_threads.append(
            threading.Thread(name="scheduler", target=start_scheduler)
        )
        for t in background_threads:
            log.info("Starting thread for {}".format(t))
            t.start()

    from bookied.web import app

    app.config["BOOKIE_PROPOSER"] = proposer
    app.config["BOOKIE_APPROVER"] = approver

    app.run(host=host, port=port, debug=debug)


@main.command()
@click.argument("queue", default="default")
def worker(queue):
    """ Start the (redis queue) worker to deal with the received messages
    """
    from .redis_con import get_redis
    from . import work

    work.unlock()
    # Let's drop the redis database and refill it from incident store
    with Connection(get_redis()):

        q = Queue("failed")
        if q.count > 0:
            log.info("Emptying Redis Queue (failed) ...")
            q.empty()
            job = q.dequeue()
            while job is not None:
                log.info("Canceling " + str(job))
                job.cancel()
                job.delete()
                job = q.dequeue()

        q = Queue("default")
        count_before = q.count

        if q.count > 0:
            log.info("Emptying Redis Queue (default) ...")
            q.empty()
            job = q.dequeue()
            while job is not None:
                log.info("Canceling " + str(job))
                job.cancel()
                job.delete()
                job = q.dequeue()
            log.info("Redis Queue cleared (jobs before=" + str(count_before) + ")")
        else:
            log.info("Empty Redis Queue initialized")

    # This runs the Worker as thread
    with Connection(get_redis()):
        w = Worker([queue])
        w.work()


@main.command()
@click.option("--proposer", default=config.get("BOOKIE_PROPOSER"))
@click.option("--approver", default=config.get("BOOKIE_APPROVER"))
def approve(proposer, approver):
    """ Send an approve ticket to the queue
    """
    from .redis_con import get_redis
    from . import work

    redis = get_redis()
    work.unlock()
    use_connection(redis)
    q = Queue(connection=redis)
    q.enqueue(work.approve, args=(), kwargs=dict(proposer=proposer, approver=approver))


@click.option("--proposer")
@click.option("--approver")
@main.command()
def scheduler(proposer, approver):
    """ Test for postponed incidents and process them
    """
    """
    from bos_incidents import factory
    from .schedule import check_scheduled
    storage = factory.get_incident_storage()
    check_scheduled(storage, func_callback=print)
    """
    from .schedule import scheduler

    scheduler(proposer=proposer, approver=approver)


if __name__ == "__main__":
    main()



import pkg_resources
from rq import use_connection, Connection, Queue
from flask import Flask, request, jsonify

from . import work
from .config import loadConfig
from .redis_con import get_redis
from .utils import resolve_hostnames
from bos_incidents import factory, exceptions
from bos_incidents.validator import IncidentValidator, InvalidIncidentFormatException
from bookiesports.normalize import IncidentsNormalizer, NotNormalizableException

from .log import log
from . import INCIDENT_CALLS

config = loadConfig()

# Flask app and parameters
app = Flask(__name__)
redis = get_redis()
use_connection(redis)

# Invident Storage
storage = factory.get_incident_storage()

validator = IncidentValidator()

# API whitelist
api_whitelist = resolve_hostnames(config.get("api_whitelist", ["0.0.0.0"]))

background_threads = []


@app.route("/")
def home():
    """ Let's not expose that this is a bos-auto endpoint
    """
    return "", 404


@app.route("/isalive")
def isalive():

    versions = dict()
    for name in ["bos-auto", "peerplays", "bookiesports", "bos-incidents", "bos-sync"]:
        try:
            versions[name] = pkg_resources.require(name)[0].version
        except Exception:
            versions[name] = "not installed"
    # queue status
    queue_status = {}
    with Connection(redis):
        q = Queue(connection=redis)
        for queue in q.all():
            queue_status[queue.name] = dict(count=queue.count)

    background_threads_dict = {}
    for t in background_threads:
        try:
            background_threads_dict[t.name] = {"running": t.is_alive()}
        except Exception as e:
            log.error("Error in background task: {}".format(str(e)))

    return jsonify(
        dict(
            versions=versions,
            queue=dict(status=queue_status),
            background=background_threads_dict,
        )
    )


@app.route("/events")
@app.route("/events/<event_id>")
def events(event_id=None):
    if event_id is not None:
        event = storage.get_event_by_id(event_id, True, True)
        event.pop("_id")
        return jsonify(dict(event))
    else:
        return jsonify(storage.get_incidents_count())


@app.route("/finalize")
def finalize():
    # only added for debugging
    for call in INCIDENT_CALLS:
        log.info("- querying call for finalizing {}".format(call))
        events = []
        for status_name in [
            "connection lost",
            "unknown",
            "postponed",
            "unhandled exception, retrying soon",
        ]:
            for event in storage.get_events_by_call_status(
                call=call, status_name=status_name
            ):
                events.append(event)

        log.info("Finalizing " + str(len(events)) + " events ...")
        for event in events:
            storage.update_event_status_by_id(
                event["id_string"], call=call, status_name="manually finalized"
            )


@app.route("/finalize/purge")
def finalize_purge():
    storage = factory.get_incident_storage(purge=True)
    return "success"


@app.route("/trigger", methods=["GET", "POST"])
def trigger():
    """ This endpoint is used to submit data to the queue so we can process it
        asynchronously to the web requests. The webrequests should be answered
        fast, while the processing might take more time

        The endpoint opens an API according to the ``--port`` and ``--host``
        settings on launch. Thise API provides an endpoint on

            /trigger

        and consumes POST messages with JSON formatted body.

        The body is validated against the incident schema defined in
        bos-incidents

        .. note:: The trigger endpoint stores the incidents through
                  (bos-incidents) already to allow later replaying.
    """
    if request.method == "POST":
        # Don't bother wit requests from IPs that are not
        # whitelisted
        if request.remote_addr not in api_whitelist and "0.0.0.0" not in api_whitelist:
            return "Your IP address is not allowed to post here!", 403

        # Obtain message from request body
        incident = request.get_json()

        # Ensure it is json
        try:
            validator.validate_incident(incident)
        except InvalidIncidentFormatException:
            log.error("Received invalid request: {}".format(str(incident)))
            return "Invalid data format", 400

        # Only accept normalizable incidents
        # Normalize incident
        normalizer = IncidentsNormalizer(chain=config.get("network", "beatrice"))
        try:
            incident = normalizer.normalize(incident, True)
        except NotNormalizableException:
            log.warning(
                "Received not normalizable incident, discarding {}".format(
                    str(incident)
                )
            )
            return "Not normalized incident", 400

        try:
            # FIXME, remove copy()
            storage.insert_incident(incident.copy())
        except exceptions.DuplicateIncidentException as e:
            # We merely pass here since we have the incident already
            # alerting anyone won't do anything
            # traceback.print_exc()
            pass

        # Send incident to redis
        with Connection(redis):
            q = Queue(connection=redis)
            job = q.enqueue(
                work.process,
                args=(incident,),
                kwargs=dict(
                    proposer=app.config.get("BOOKIE_PROPOSER"),
                    approver=app.config.get("BOOKIE_APPROVER"),
                ),
            )
            log.info("Forwarded incident {} to worker via redis".format(str(incident)))

            # In case we "proposed" something, we also need to approve,
            # we do that by queuing a approve
            approve_job = q.enqueue(
                work.approve,
                args=(),
                kwargs=dict(
                    proposer=app.config.get("BOOKIE_PROPOSER"),
                    approver=app.config.get("BOOKIE_APPROVER"),
                ),
            )

        # Return message with id
        return jsonify(
            dict(
                result="processing",
                message=incident,
                id=str(job.id),
                id_approve=str(approve_job.id),
            )
        )

    return "", 503



from .trigger import Trigger
from ..log import log
from .. import exceptions


class FinishTrigger(Trigger):
    """ The finish trigger merely sets an event to finished on whistle time
    """

    def _trigger(self, args):
        """ Set a BMG to ``finish``.
        """
        log.info(
            "Finishing an event by setting it to 'finished'" " (without results)..."
        )

        event = self.getEvent()

        event.status_update("finished", scores=[])

        return True

    def testThreshold(self):
        """ The threshold that needs to be crossed in order to finish an event
            on chain.

            .. alert:: This is temporary set to be ``2`` until we have an
                easier way to identify how many data proxies send data to us
        """
        return 2

    def testConditions(self, *args, **kwargs):
        """ The test conditions for finishing the event are as this:

            * Do more incidents propose the creation of the event than
                ``testThreshold``

               -> Finish the event

        """
        incidents = self.get_all_incidents()
        if not incidents:
            raise exceptions.InsufficientIncidents
        finish_incidents = incidents.get("finish", {}).get("incidents", [])
        if not finish_incidents:
            raise exceptions.InsufficientIncidents
        provider_hashes = set()
        for incident in finish_incidents:
            provider_hash = incident.get("provider_info", {}).get("name", None)
            if provider_hash is not None:
                provider_hashes.add(provider_hash)
        if len(provider_hashes) >= self.testThreshold():
            return True
        else:
            log.info(
                "Insufficient incidents for {}({})".format(
                    self.__class__.__name__, str(self.teams)
                )
            )
            raise exceptions.InsufficientIncidents



class EventDoesNotExistException(Exception):
    """ An event does not exist
    """

    pass


class EventGroupClosedException(Exception):
    """ The event group is closed and no events can open
    """

    pass


class ProposalAlreadyExistsOrIsPendingException(Exception):
    """ The event group is closed and no events can open
    """

    pass


class EventCannotOpenException(Exception):
    """ An event cannot be opened yet, possibly due to
        leadtime_Max setting
    """

    pass


class InsufficientIncidents(Exception):
    """ Insufficient incidents to make a decision
    """

    pass


class InsufficientEqualResults(Exception):
    """ Inconsistent result incidents that don't allow
        us to make a decision
    """

    pass


class TooManyDifferentResultsOverThreshold(Exception):
    """ To many different results appear to be above the required threshold.
        This should never have happend unless the thresholds are too low
    """

    pass


class CreateIncidentTooOldException(Exception):
    """ The create incident tries to create an event in the past
    """

    pass


class InvalidParticipantException(Exception):
    pass


class PostPoneIncidentException(Exception):
    pass



from .trigger import Trigger
from ..log import log


class InProgressTrigger(Trigger):
    """ The in progress trigger sets an event to in_play as soon as it starts
    """

    def _trigger(self, args):
        """ Set a BMG to ``in_progress``
        """
        log.info("Setting a event to 'in_progress'...")
        event = self.getEvent()
        event.status_update("in_progress")

    def testConditions(self, *args, **kwargs):
        """ The conditions are always true as we want the
            event to be opened as soon as possible
        """
        return True



import os
import yaml


def loadConfig(config_file="config.yaml"):
    """ Load the configuration from a YAML file
    """
    config = dict()
    default_file = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "config-default.yaml"
    )
    with open(default_file) as fid:
        config.update(yaml.load(fid))

    # Load wallet passphrase from config
    if os.path.isfile(config_file):
        with open(config_file, "r") as fid:
            config.update(yaml.load(fid))

    return config



from . import SKIP_DYNAMIC_BMS
from .trigger import Trigger
from .. import exceptions
from ..log import log
from bookied_sync.event import LookupEvent
from datetime import datetime


class CreateTrigger(Trigger):
    """ This trigger inherits class:`Trigger` and deals with create incidents
        that are fired by the data proxy when new events are announced.
    """

    def _trigger(self, args):
        """ Trigger the 'create' message
        """
        log.info("Creating a new event ...")

        # Let's see if we can find an Event in the Lookup
        self.event = self.getIncidentEvent()
        if not self.event:
            return

        # Set parameters
        season = args.get("season")
        if isinstance(season, str):
            season = {"en": season}
        if self.event["season"] and not (
            self.event["season"].get("en") in season.get("en")
            or season.get("en") in self.event["season"].get("en")
        ):
            err = "Seasons don't match: {} != {}".format(
                season.get("en"), self.event["season"].get("en")
            )
            log.critical(err)
            raise Exception(err)
        self.event["season"] = season
        self.event["status"] = "upcoming"

        # Update self.event
        self.event.update()

        # Create Betting market Groups
        self.createBmgs()

        return True

    def createBmgs(self):
        """ Go through all Betting Market groups and create them
        """
        for bmg in self.event.bettingmarketgroups:

            # Skip dynamic bmgs
            if bmg["dynamic"]:
                # Dynamic BMGs are created separately
                log.debug("Skipping dynamic BMG: {}".format(str(bmg.identifier)))
                continue
            bmg.update()
            self.createBms(bmg)

    def createBms(self, bmg):
        """ Go through all betting markets and create them
        """
        log.debug("Updating Betting Markets ...")
        for bm in bmg.bettingmarkets:
            log.debug(
                "Updating Betting Market {} ...".format(bm["description"].get("en"))
            )
            bm.update()

    def getIncidentEvent(self):
        """ Does not throw in all cases but returns None in case of error
        """
        # Obtain event
        try:
            return self.getEvent()
        except exceptions.EventDoesNotExistException:
            try:
                log.info(
                    "Creating event with teams {} in group {}.".format(
                        str(self.teams), self.eventgroup.identifier
                    )
                )
                # Let's create the event in this case!
                return self.createEvent()
            except exceptions.EventCannotOpenException as e:
                msg = "The event with teams {} in group {} cannot open yet: {}".format(
                    str(self.teams), self.eventgroup.identifier, str(e)
                )
                log.info(msg)
                raise exceptions.EventCannotOpenException(msg)

    def createEvent(self):
        """ Create event
        """
        self.event = LookupEvent(
            teams=self.teams,
            start_time=self.start_time,
            eventgroup_identifier=self.eventgroup.identifier,
            sport_identifier=self.sport.identifier,
        )

        # This tests for leadtime_max
        if not self.event.can_open:
            can_open_by = self.event.can_open_by
            self.set_incident_status(
                status_name="postponed", status_expiration=can_open_by
            )
            raise exceptions.EventCannotOpenException(
                "Can only open after {}".format(str(can_open_by))
            )

        return self.event

    def testThreshold(self):
        """ The threshold that needs to be crossed in order to create an event
            on chain.

            .. alert:: This is temporary set to be ``2`` until we have an
                easier way to identify how many data proxies send data to us

        """
        return 2

    def testConditions(self, *args, **kwargs):
        """ The test conditions for creating the event are as this:

            * Do more incidents propose the creation of the event than
                ``testThreshold``

               -> Create the event

        """

        # This case can be triggered early and does not need
        # to wait for when we actually try to create an event
        if self.start_time.replace(tzinfo=None) < datetime.utcnow():
            raise exceptions.CreateIncidentTooOldException(
                "Won't create event. " "'start_time' was {}".format(self.start_time)
            )

        # Do not create event that has to few incidents
        incidents = self.get_all_incidents()
        if not incidents:
            raise exceptions.InsufficientIncidents("No incident found")
        create_incidents = incidents.get("create", {}).get("incidents", [])
        provider_hashes = set()
        for incident in create_incidents:
            provider_hash = incident.get("provider_info", {}).get("name", None)
            if provider_hash is not None:
                provider_hashes.add(provider_hash)
        if len(provider_hashes) >= self.testThreshold():
            return True
        else:
            msg = "Insufficient incidents for {}({})".format(
                self.__class__.__name__, str(self.teams)
            )
            log.info(msg)
            raise exceptions.InsufficientIncidents(msg)



import threading
import time
from datetime import datetime

from rq import Queue
from bos_incidents import factory

from . import INCIDENT_CALLS
from .log import log
from .config import loadConfig
from .redis_con import get_redis

config = loadConfig()


def check_scheduled(
    storage=None, func_callback=None, proposer=None, approver=None
):  # pragma: no cover
    """
    """
    log.info(
        "Scheduler checking incidents database ... "
        "(approver: {}/ proposer: {})".format(approver, proposer)
    )

    # Invident Storage
    if not storage:
        storage = factory.get_incident_storage()

    if not proposer:
        proposer = config.get("BOOKIE_PROPOSER")
    if not approver:
        approver = config.get("BOOKIE_APPROVER")

    push_to_queue = []

    for call in INCIDENT_CALLS:
        log.info("- querying call {}".format(call))

        events = []

        for status_name in ["postponed", "unhandled exception, retrying soon"]:
            for event in storage.get_events_by_call_status(
                call=call,
                status_name=status_name,
                status_expired_before=datetime.utcnow(),
            ):
                events.append(event)

        for status_name in ["connection lost", "unknown"]:
            for event in storage.get_events_by_call_status(
                call=call, status_name=status_name
            ):
                events.append(event)

        ids = list()
        log.info("Scheduler retriggering " + str(len(events)) + " incident ...")
        for event in events:
            for incidentid in event.get(call, {}).get("incidents", []):
                incident = storage.resolve_to_incident(incidentid)
                if func_callback:
                    push_to_queue.append(incident)
                    # it is enough to trigger one incident, worker will check the whole call
                    break

    # Flask queue
    q = Queue(connection=get_redis())

    # only push into the queue if it's somewhat empty (with 10% buffer), otherwise wait
    if q.count + 2 < len(push_to_queue):
        for incident in push_to_queue:
            job = q.enqueue(
                func_callback,
                args=(incident,),
                kwargs=dict(proposer=proposer, approver=approver),
            )
            ids.append(job.id)

    return ids


class PeriodicExecutor(threading.Thread):  # pragma: no cover
    """ Periodically execute a task
    """

    def __init__(self, sleep, func, *args, **kwargs):

        """ execute func(params) every 'sleep' seconds """
        self.func = func
        self.sleep = sleep
        self.args = args
        self.kwargs = kwargs
        threading.Thread.__init__(self, name="PeriodicExecutor")
        self.setDaemon(True)

    def run(self):
        while True:
            try:
                self.func(*self.args, **self.kwargs)
                time.sleep(self.sleep)
            except Exception as e:
                log.error(str(e))
                time.sleep(self.sleep)


def scheduler(delay=None, proposer=None, approver=None):  # pragma: no cover
    """
    """
    from . import work

    if not delay:
        delay = config["scheduler"]["interval"]

    if not proposer:
        proposer = config.get("BOOKIE_PROPOSER")
    if not approver:
        approver = config.get("BOOKIE_APPROVER")

    check_scheduled(
        storage=None, func_callback=work.process, proposer=proposer, approver=approver
    )

    PeriodicExecutor(
        delay,
        check_scheduled,
        storage=None,
        func_callback=work.process,
        proposer=proposer,
        approver=approver,
    ).run()



import time
import logging
import traceback
import grapheneapi
import bookied_sync
import bos_incidents
from bos_incidents.format import id_to_string
import random

from flask_rq import job
from datetime import datetime, timedelta

from bookied_sync.lookup import Lookup
from peerplays import PeerPlays
from peerplays.instance import set_shared_blockchain_instance
from peerplays.account import Account
from peerplays.proposal import Proposals

from bookiesports.normalize import NotNormalizableException

from .log import log
from .config import loadConfig
from . import exceptions, TRIGGERS


config = loadConfig()
peerplays = PeerPlays(
    node=config.get("node", None),
    nobroadcast=config.get("nobroadcast", False),
    num_retries=1,  # Only try once then trow an exception
)
set_shared_blockchain_instance(peerplays)
lookup = Lookup(
    proposing_account=config.get("BOOKIE_PROPOSER"),
    approving_account=config.get("BOOKIE_APPROVER"),
    blockchain_instance=peerplays,
    network=config.get("network", "beatrice"),
)


def unlock():
    """ Unlock the python-peerplays wallet so we can sign the things we are
        proposing/approving.
    """
    # We need to know the passphrase to unlock the wallet
    if "passphrase" not in config:
        err = "No 'passphrase' found in configuration!"
        log.critical(err)
        raise ValueError(err)

    if not lookup.wallet.created():
        err = "Please create a wallet and import the keys first!"
        log.critical(err)
        raise Exception(err)

    print("unlocking wallet ...")
    lookup.wallet.unlock(config.get("passphrase"))


#
# Processing Job
#
@job
def process(message, **kwargs):
    """ This process is called by the queue to process an actual message
        received.

        Hence, this method has the look and feel of a dispatcher!
    """
    # We clear the buffers here so we start fresh with no
    # operation in any buffer
    lookup.clear()

    # Let's also reconnect
    peerplays.rpc.connect()
    try:
        t = time.time()
        log.info(
            "processing "
            + message["unique_string"]
            + " for event "
            + id_to_string(message)
        )
    except Exception as e:
        pass

    assert isinstance(message, dict)
    assert "id" in message

    approver = kwargs.get("approver", None)
    if not approver:
        approver = message.get("approver", None)
    if approver:
        lookup.set_approving_account(approver)

    proposer = kwargs.get("proposer", None)
    if not proposer:
        proposer = message.get("proposer", None)
    if proposer:
        lookup.set_proposing_account(proposer)

    log.info("Proposer account: {}".format(lookup.proposing_account))
    log.info("Approver account: {}".format(lookup.approving_account))

    # Call
    call = message.get("call").lower()

    # Obtain arguments
    args = message.get("arguments")

    log.info("initializing {} call with args {}".format(call, str(args)))
    try:

        if call in TRIGGERS:
            trigger = TRIGGERS[call](message, lookup_instance=lookup, config=config)

        elif call == "unknown":
            return

        else:
            log.error(
                "Received an unknown trigger {} with content: {}".format(call, message)
            )
            return

    except bookied_sync.exceptions.ObjectNotFoundInLookup as e:
        log.info(str(e))
        return

    except NotNormalizableException as e:
        log.warning("Incident not normalizable: {}".format(message))
        return

    except Exception as e:
        log.critical(
            "Uncaught exception: {}\n\n{}".format(str(e), traceback.format_exc())
        )
        # No trigger can be executed!
        return

    log.info("processing {} call from event id {}".format(call, str(trigger.id)))

    try:
        # Execute the trigger
        trigger.trigger(args)

    except exceptions.EventDoesNotExistException:
        log.warning("Event does not exist on chain!")
        trigger.set_incident_status(status_name="event missing")

    except exceptions.EventGroupClosedException:
        log.warning("The event group is closed!")
        trigger.set_incident_status(status_name="event group closed")

    except exceptions.EventCannotOpenException:
        log.info("This incident has been postponed")
        """
        now = int(time.mktime(datetime.utcnow().timetuple()))
        expiration_time = trigger.event.can_open_by + timedelta(seconds=(now % 120))

        trigger.set_incident_status(
            status_name="postponed",
            status_expiration=expiration_time
        )

        """
        trigger.set_incident_status(
            status_name="postponed",
            status_expiration=datetime.utcnow()
            + timedelta(
                seconds=int(
                    config["scheduler"]["expirations"]["EventCannotOpenException"]
                )
            ),
        )

    except exceptions.PostPoneIncidentException:
        log.info("This incident has been postponed")
        trigger.set_incident_status(
            status_name="postponed",
            status_expiration=datetime.utcnow()
            + timedelta(
                seconds=int(
                    config["scheduler"]["expirations"]["PostPoneIncidentException"]
                )
            ),
        )

    except exceptions.InsufficientIncidents:
        trigger.set_incident_status(status_name="insufficient incidents")

    except exceptions.InsufficientEqualResults:
        log.info("This incident couldn't lead to a decision")
        trigger.set_incident_status(status_name="undecided")

    except grapheneapi.exceptions.NumRetriesReached:
        log.critical("Connection to backend node has been lost!")
        trigger.set_incident_status(status_name="connection lost")

    except bos_incidents.exceptions.EventNotFoundException as e:
        # We do not set the incident status because this requires an event to
        # exist which doesn't - that's why we are in here, afterall
        # trigger.set_incident_status(status_name="event missing in bos_incidents")
        log.info("Invalid bos_incident event!")

    except bookied_sync.exceptions.ObjectNotFoundInLookup as e:
        trigger.set_incident_status(status_name="related object not found")
        log.info(str(e))

    except bookied_sync.exceptions.ObjectNotFoundError as e:
        trigger.set_incident_status(status_name="related object not found")
        log.info(str(e))

    except exceptions.CreateIncidentTooOldException as e:
        trigger.set_incident_status(status_name="create too old")
        log.warning(str(e))

    except Exception as e:
        exception_details = "{}\n\n{}".format(str(e), traceback.format_exc())

        # retry uncaught exception once (to reduce ghost errors)
        if trigger.get_incident_status() is not None and (
            trigger.get_incident_status()["name"]
            == "unhandled exception, retrying soon"
            or trigger.get_incident_status()["name"] == "unhandled exception"
        ):
            # already retried, finalize
            trigger.set_incident_status(
                status_name="unhandled exception",
                status_add={"details": exception_details},
            )
            log.critical(
                "Uncaught exception: {}\n\n{}".format(str(e), traceback.format_exc())
            )
        else:
            # randomize it
            random_offset = random.randint(1, 3) * 60
            expiration_in_seconds = max(
                30,
                int(config["scheduler"]["expirations"]["UnhandledException"])
                - random_offset,
            )

            trigger.set_incident_status(
                status_name="unhandled exception, retrying soon",
                status_expiration=datetime.utcnow()
                + timedelta(seconds=expiration_in_seconds),
                status_add={"details": exception_details},
            )
            log.info("Uncaught exception, retrying soon: {}".format(str(e)))

    try:
        elapsed = time.time() - t
        log.info(
            "Done processing "
            + message["unique_string"]
            + ", call status now "
            + trigger.get_incident_status()
            + ", elapsed time is "
            + str(elapsed)
        )
    except Exception as e:
        pass


#
# Approve my own Proposals
#
@job
def approve(*args, **kwargs):
    """ This process is meant to approve proposals that I have created.

        The reason for this is that proposals created by accountA are not
        automatically also approved by accountA and need an explicit approval.
    """
    peerplays.rpc.connect()
    myapprover = kwargs.get("approver", None)
    if not myapprover:
        myapprover = config.get("BOOKIE_APPROVER")

    myproposer = kwargs.get("proposer", None)
    if not myproposer:
        myproposer = config.get("BOOKIE_PROPOSER")

    log.info(
        "Testing for pending proposals "
        "created by {} that we could approve by {}".format(myproposer, myapprover)
    )

    proposals = Proposals("witness-account")
    approver = Account(myapprover)
    for proposal in proposals:
        proposer = Account(proposal.proposer)
        if (
            proposer["name"] == myproposer
            and approver["id"] not in proposal["available_active_approvals"]
        ):
            log.info(
                "Proposal {} has been proposed by {}. Approving it!".format(
                    proposal["id"], myproposer
                )
            )
            log.info(peerplays.approveproposal(proposal["id"], account=myapprover))



import math
import statistics

from datetime import datetime, timedelta
from dateutil.parser import parse

from bookied_sync import comparators
from bookied_sync.event import LookupEvent
from bookied_sync.bettingmarketgroup import LookupBettingMarketGroup
from bookied_sync.bettingmarket import LookupBettingMarket

from .trigger import Trigger
from .. import exceptions
from ..log import log
from ..utils import dList2Dict
from ..config import loadConfig

MIN_AGE_INCIDENT = 1


config = loadConfig()


def obtain_participant_side(participant, teams):
    if participant not in teams:
        raise exceptions.InvalidParticipantException
    return "home" if teams.index(participant) == 0 else "away"


class DynamicBmgTrigger(Trigger):
    """ This trigger inherits class:`Trigger` and deals with the dynamic_bmgs
        trigger.
    """

    def _trigger(self, args):
        """ Trigger the 'dynamic' message
        """
        self.event = self.getEvent()
        self.teams = self.event.teams

        on_chain_bmgs = self.get_onchain_bmgs()
        incident_types = self.incident["arguments"]["types"]

        # We need to deal with each of the incidents individually
        log.debug("Testing incidents and existing BMGs")
        log.debug("Number of types in the incident:  {}".format(len(incident_types)))
        log.debug("Number of existing on-chain BMGs: {}".format(len(on_chain_bmgs)))

        types_done = list()
        for incident_type in incident_types:
            # This type of incident has multiple incident types, we go through
            # all of them individually

            # Let's see if we can find the corresponding BMG on chain already
            exists_on_chain = False
            for chain_bmg in on_chain_bmgs:

                # Test if we may have a dynamic BMG of that group on chain already:
                if chain_bmg.is_dynamic() and chain_bmg.is_dynamic_type(
                    incident_type["type"]
                ):
                    exists_on_chain = True
                    # We *DO* have this on chain already!
                    log.info(
                        "A dynamic BMG of type {} exists for event {} already".format(
                            incident_type["type"], self.event.identifier
                        )
                    )

            # Let's skip if we have we this type exists on chain already, or
            # we've already worked on one that is similar
            if exists_on_chain or incident_type["type"] in types_done:
                log.info("Already found BMGs on chain.. Aborting")
                continue

            # ... else, we create the BMG and corresponding BMs
            log.info("Creating dynamic BMGs ...")
            try:
                self.createBmg(self.event, incident_type)
            except Exception as e:
                # If an exception is raised, let's log and continue
                log.critical("{}: {}".format(e.__class__.__name__, str(e)))

            types_done.append(incident_type["type"])

    def createBmg(self, event, incident_type):
        """ Go through all Betting Market groups and create the one that
            matches the type in the incident together with the corresponding
            betting markets.

            This method is called for each type in the incident, that means, if
            an incident contains a handicap and an over under, this method will
            be called
        """
        assert "type" in incident_type
        assert "value" in incident_type
        typ = incident_type["type"]

        # Let's find the BM according to bookiesports
        bmgs = list(event.bettingmarketgroups)
        log.debug("Expected number of BMGs: {}".format(len(bmgs)))
        for bmg in bmgs:

            # Only do dynamic ones here
            if not bmg.get("dynamic"):
                log.debug("BMG is not dynamic: {}".format(bmg.identifier))
                continue

            # We only deal with the one corresponding with the incident_type
            if not LookupBettingMarketGroup.is_dynamic_type(bmg["dynamic"], typ):
                log.debug("BMG {} is not of type {}".format(bmg.identifier, typ))
                continue

            # If incident is a Overunder BMG
            if LookupBettingMarketGroup.is_ou_type(
                bmg.get("dynamic")
            ) and LookupBettingMarketGroup.is_ou_type(typ):
                self.updateOverUnderBMg(bmg)
                self.createBms(bmg)
                return

            # If incident is a Handicap BMG
            elif LookupBettingMarketGroup.is_hc_type(
                bmg.get("dynamic")
            ) and LookupBettingMarketGroup.is_hc_type(typ):
                self.updateHandicapBMg(bmg, incident_type)
                self.createBms(bmg)
                return

            else:
                log.error(
                    "BMG is could not be classified: {} - Type: {}".format(
                        bmg.identifier, (typ or "empty string")
                    )
                )

    def updateOverUnderBMg(self, bmg):
        """ This method creates a BMG for OverUnder
        """
        log.debug("BMG is dynamic Over/Under: {}".format(bmg.identifier))

        # Let's obtain the overunder value
        overunder = self.median_value("ou")

        # Set Overunder
        bmg.set_overunder(overunder)

        # We use fuzzy arguments by defining and injecting our own
        # test_operation_equal_search and find_id_search comparators.
        # This allows us to select what to actually look for.
        fuzzy_args = {
            "test_operation_equal_search": [
                # The following keys are required
                comparators.cmp_required_keys(
                    [  # other for updates with new_*
                        "betting_market_group_id",
                        "new_description",
                        "new_event_id",
                        "new_rules_id",
                    ],
                    [  # or as actual creates
                        "betting_market_group_id",
                        "description",
                        "event_id",
                        "rules_id",
                    ],
                ),
                # The status needs to match, if not update
                comparators.cmp_status(),
                # The status needs to match
                comparators.cmp_event(),
                # The description must have a _dynamic attribute
                comparators.cmp_description("_dynamic"),
                # The fuzzy comparator takes the fuzzyness from config
                # and tries to find the bmgs for handicap and over
                # under dynamic factors but allows some fuzzyness for
                # the actual value
                comparators.cmp_dynamic_bmg_fuzzy(
                    config["dynamic"]["overunder"]["fuzzy_value"]
                ),
            ],
            # We identify the bmg by comparing
            "find_id_search": [
                # parent event id
                comparators.cmp_event(),
                # comparing type of dynamic
                comparators.cmp_description("_dynamic"),
                # fuzzy matching of dynamic value
                comparators.cmp_dynamic_bmg_fuzzy(
                    config["dynamic"]["overunder"]["fuzzy_value"]
                ),
            ],
        }
        bmg.update(**fuzzy_args)

    def updateHandicapBMg(self, bmg, incident_type):
        """ This method creates a BMG for Hanidcap
        """
        log.debug("BMG is dynamic Handicap: {}".format(bmg.identifier))

        # Identify which player has the handicap
        side = obtain_participant_side(incident_type["participant"], self.teams)

        # Set handicap
        handicap = round(self.median_value("hc", side=side))
        bmg.set_handicaps(**{side: handicap})

        # We use fuzzy arguments by defining and injecting our own
        # test_operation_equal_search and find_id_search comparators.
        # This allows us to select what to actually look for.
        fuzzy_args = {
            "test_operation_equal_search": [
                # The following keys are required
                comparators.cmp_required_keys(
                    [  # other for updates with new_*
                        "betting_market_group_id",
                        "new_description",
                        "new_event_id",
                        "new_rules_id",
                    ],
                    [  # or as actual creates
                        "betting_market_group_id",
                        "description",
                        "event_id",
                        "rules_id",
                    ],
                ),
                # The status needs to match, if not update
                comparators.cmp_status(),
                # The status needs to match
                comparators.cmp_event(),
                # The description must have a _dynamic attribute
                comparators.cmp_description("_dynamic"),
                # The fuzzy comparator takes the fuzzyness from config
                # and tries to find the bmgs for handicap and over
                # under dynamic factors but allows some fuzzyness for
                # the actual value
                comparators.cmp_dynamic_bmg_fuzzy(
                    config["dynamic"]["handicap"]["fuzzy_value"]
                ),
            ],
            # We identify the bmg by comparing
            "find_id_search": [
                # parent event id
                comparators.cmp_event(),
                # comparing type of dynamic
                comparators.cmp_description("_dynamic"),
                # fuzzy matching of dynamic value
                comparators.cmp_dynamic_bmg_fuzzy(
                    config["dynamic"]["handicap"]["fuzzy_value"]
                ),
            ],
        }
        bmg.update(**fuzzy_args)

    def createBms(self, bmg):
        """ Go through all betting markets and create them
        """
        log.debug("Updating Betting Markets ...")
        for bm in bmg.bettingmarkets:
            log.debug(
                "Updating Betting Market {} ...".format(bm["description"].get("en"))
            )
            bm.update()

    def testConditions(self, *args, **kwargs):
        """ The test conditions for creating the event are as this:

            For this trigger, the threshold is based on time. After receiving
            the first incident, we wait time X (config) per type. If passed, we
            continue, else we only store.

        """
        # We test if the incident we received is X hours old
        # This happens if the incident is retriggered after being postpond
        my_time = parse(self.incident["timestamp"]).replace(tzinfo=None)
        time_limit = datetime.utcnow() - timedelta(
            seconds=config["conditions"]["dynamic_bmgs"]["incident-min-age"]
        )

        # We only allow this trigger if it is older than x hours
        if my_time <= time_limit:
            return True

        else:
            msg = "Incident not old enough {}({})".format(
                self.__class__.__name__, str(self.teams)
            )
            log.info(msg)
            raise exceptions.PostPoneIncidentException(msg)

    def median_value(self, type, side=None):
        """ This method is used to obtain the median value of all incidents
            that provide a value for a certain incident type. This ensure that
            multiple incident providers with different opinion about the
            dynamic market will result in just the median to be created.
        """
        log.debug("Obtaining median value of incidents")

        # Get all incidents for the current event
        incidents = self.get_all_incidents()

        # We raise if no incident is found, this should really not happen.
        # Also, there should be multiple incidents matching this query
        # according to testConditions()
        if not incidents:
            raise exceptions.InsufficientIncidents("No incident found")

        # Let's resolve the dynamic incidents
        dynamic_incidents = incidents.get("dynamic_bmgs", {}).get("incidents", [])
        incidents = [self.storage.resolve_to_incident(x) for x in dynamic_incidents]

        # Obtain those incidents that have our type
        values = list()
        for incident in incidents:

            for incident_type in incident["arguments"]["types"]:

                # Same type?
                if not LookupBettingMarketGroup.is_dynamic_type(
                    incident_type["type"], type
                ):
                    # .. of skip, if not
                    continue

                # Let's see if this is handicap type
                if LookupBettingMarketGroup.is_hc_type(type):
                    log.debug(
                        "Dealing with Handicap value: {}".format(incident_type["value"])
                    )

                    # Correct team orientation?
                    this_side = obtain_participant_side(
                        incident_type["participant"], self.teams
                    )
                    # If sides don't match, we need to invert values
                    if this_side == side:
                        values.append(float(incident_type["value"]))
                    else:
                        values.append(-float(incident_type["value"]))

                # let's see if this is over under type
                elif LookupBettingMarketGroup.is_ou_type(type):
                    log.debug(
                        "Dealing with over/under value: {}".format(
                            incident_type["value"]
                        )
                    )

                    # store value
                    values.append(float(incident_type["value"]))

        if values:
            log.debug("Obtaining median ({}) for values: {}".format(type, str(values)))
            return statistics.median(values)
        else:
            log.warning("No values could be processed for type {}!".format(type))



from .trigger import Trigger
from bookied_sync import comparators
from bookied_sync.bettingmarketgroup import LookupBettingMarketGroup
from bookied_sync.bettingmarketgroupresolve import LookupBettingMarketGroupResolve
from ..log import log
from .. import exceptions


#: Internal score separator
_SCORE_SEPARATOR = "::"


class ResultTrigger(Trigger):
    """ This trigger deals with the results of an event and needs to combine
        information from different incident reports.
    """

    def _trigger(self, args):
        """ Publish results to a BMG and set to ``finish``.
        """
        log.info("Finishing an event by setting it to " "'finished' (with results)...")

        event = self.getEvent()

        event.status_update(
            "finished", scores=[str(self.home_score), str(self.away_score)]
        )

        log.info("Settling betting market groups...")

        self.resolve_bgms(event)

    def resolve_bgms(self, event):
        """ Resolve the BMGs
        """
        for bmg in event.bettingmarketgroups:
            if bmg["dynamic"]:
                # Update and crate BMs
                fuzzy_args = {
                    "test_operation_equal_search": [
                        comparators.cmp_event(),
                        comparators.cmp_description("_dynamic"),
                    ],
                    "find_id_search": [
                        comparators.cmp_event(),
                        comparators.cmp_description("_dynamic"),
                    ],
                }
            else:
                fuzzy_args = {}

            # Let's try find an id with fuzzy
            # This also sets the dynamic parameters in the bmg object!
            _ = bmg.find_id(**fuzzy_args)

            # Skip those bmgs that coudn't be found
            if not bmg.find_id():
                log.error("BMG could not be found: {}".format(str(bmg.identifier)))
                continue

            settle = LookupBettingMarketGroupResolve(
                bmg, [self.home_score, self.away_score]
            )
            settle.update()

    def testThreshold(self):
        """ The threshold that needs to be crossed in order to "grade" an event
            on chain (send out the results).

            .. alert:: This is temporary set to be ``2`` until we have an
                easier way to identify how many data proxies send data to us
        """
        return 2

    def testThresholdPercentage(self):
        """ This is the percentage of incidents that needs to agree on the
            results before we push them on-chain
        """
        return 51

    def _count_score(self, scores):
        """ Internal counter of scores
        """
        from collections import Counter

        ret = Counter()
        for i in range(len(scores)):
            ret[scores[i]] += 1
        return ret

    def testConditions(self, *args, **kwargs):
        """ The test conditions for "grading" the event are as this:

            * Raise if less than ``testThreshold`` result incidents received
            * Raise if less than ``testThresholdPercentage`` result incidents agree on the result
            * Raise if multiple results reach above thresholds
            * Else, publish the result that is left

        """
        incidents = self.get_all_incidents()
        if not incidents:
            raise exceptions.InsufficientIncidents
        result_incidents = incidents.get("result", {}).get("incidents", [])
        if not result_incidents:
            raise exceptions.InsufficientIncidents
        provider_hashes = set()
        for incident in result_incidents:
            provider_hash = incident.get("provider_info", {}).get("name", None)
            if provider_hash is not None:
                provider_hashes.add(provider_hash)
        if len(provider_hashes) < self.testThreshold():
            log.info(
                "Insufficient incidents for {}({})".format(
                    self.__class__.__name__, str(self.teams)
                )
            )
            raise exceptions.InsufficientIncidents

        # Figure out the most "probable" result
        scores = [
            "{}{}{}".format(
                x["arguments"]["away_score"],
                _SCORE_SEPARATOR,
                x["arguments"]["home_score"],
            )
            for x in result_incidents
        ]
        scores = self._count_score(scores)

        # Threshold scaled by number of scores
        threshold = len(scores) * self.testThresholdPercentage() / 100

        # Valid results filtered by threshold
        valid_results = dict({k: v for k, v in scores.items() if v >= threshold})

        # Raise if multiple results are valid
        if len(valid_results) > 1:
            log.info("Too many different results over threshold")
            raise exceptions.TooManyDifferentResultsOverThreshold(valid_results)
        elif not valid_results:
            log.info("Insufficient Equal Results")
            raise exceptions.InsufficientEqualResults
        else:
            result = list(valid_results.keys())[0]
            self.away_score, self.home_score = result.split(_SCORE_SEPARATOR)
            log.info(
                "Resolving: home - away: ({} - {})".format(
                    self.home_score, self.away_score
                )
            )
            return True



#!/usr/bin/env python

from setuptools import setup
import sys

__VERSION__ = "0.4.1"

assert sys.version_info[0] == 3, "We require Python > 3"

setup(
    name="bos-auto",
    version=__VERSION__,
    description=("A daemon to manage synchronization with the blockchain"),
    long_description=open("README.md").read(),
    download_url="https://github.com/pbsa/bos-auto/tarball/" + __VERSION__,
    author="Peerplays Community",
    maintainer="Peerplays Community",
    url="http://pbsa.info",
    keywords=["peerplays", "bos"],
    packages=["bookied", "bookied.triggers"],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
    ],
    entry_points={"console_scripts": ["bos-auto = bookied.cli:main"]},
    install_requires=[open("requirements.txt").readlines()],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    include_package_data=True,
)
